;Display Test for controllerless LCD with 128x72 pixels (VG-G120751)
;ATtiny2313 @ 8 MHz, fuses: low = 0xC4, high = 0xDB
;
;PORTS
;  PB0~7 = LCD D0~D7
;  PD0   = [unused]
;  PD1   = DIO2 (first line)
;  PD2   = M/FR (toggles after each frame)
;  PD3   = DISP/DSPOF (low = display off, high = display on)
;  PD4   = CP (pixel clock)
;  PD5   = PWM output for boost converter
;  PD6   = Load (pixel data latch) & YSCLK (line clock)
;
;VERSIONS
;  2014-09-20  Arne Rossius
;    * first version

.equ	MODE = 0 ;0 = image, 1 = text
.equ	BOOST_PWM = 45 ;boost converter PWM duty cycle, max. 255



.include "../include/tn2313def.inc"

.equ	LCD_DIO2 = 1
.equ	LCD_FR = 2
.equ	LCD_ON = 3
.equ	LCD_CP = 4
.equ	LCD_LOAD = 6

.def	sreg_backup = R15

.def	temp = R16
.def	temp2 = R17
.def	count = R18
.def	line = R19
.def	scanline = R20

;===============================================================================

.cseg
.org 0x000
	rjmp	reset

.org OC1Aaddr
	rjmp	line_interrupt
	
;===============================================================================

.if (MODE == 0) ;image
image:
	.include "images/fish.asm"
.elseif (MODE == 1) ;text
text:
	.db "Lorem ipsum dolor sit", "amet, consectetur    "
	.db "adipisicing elit, sed", "do eiusmod tempor    "
	.db "incididunt ut labore ", "et dolore magna      "
	.db "aliqua. Ut enim ad   ", "minim veniam, quis   "
	.db "nostrud exercitation ",0
	
charset:
	.include "images/charset_6x9_ascii.asm"
.else
	.error "MODE must be 0 (image) or 1 (text)"
.endif

;===============================================================================

line_interrupt:
	;LCD line interrupt
	in	sreg_backup, SREG
	push	temp
	push	temp2
	push	count
	
	in	temp2, PORTD
	
	;dummy shift for unused segment outputs
	mov	temp, temp2
	ori	temp, 1<<LCD_CP
	ldi	count, (160-128)/8
line_interrupt_dummy:
	out	PORTD, temp ;CP high
	dec	count
	out	PORTD, temp2 ;CP low
	brne	line_interrupt_dummy
	
.if (MODE == 0) ;image

	;shift data (128 pixels)
	ldi	count, 128/8
line_interrupt_image:
	lpm	temp, Z+
	out	PORTB, temp ;output data
	sbi	PORTD, LCD_CP ;CP high
	dec	count
	out	PORTD, temp2 ;CP low
	brne	line_interrupt_image

.elseif (MODE == 1) ;text
	
	;shift the first 20 chars (120 px)
	ldi	count, 20/4
line_interrupt_text:
	;char 1
	movw	ZH:ZL, YH:YL
	lpm	temp, Z
	adiw	YH:YL, 1
	mov	ZL, temp
	clr	ZH
	lsl	ZL
	rol	ZH
	lsl	ZL
	rol	ZH
	lsl	ZL
	rol	ZH
	add	ZL, temp
	ldi	temp, 0
	adc	ZH, temp
	add	ZL, scanline
	adc	ZH, temp
	subi	ZL, LOW(-charset*2)
	sbci	ZH, HIGH(-charset*2)
	lpm	temp, Z
	;char 2
	movw	ZH:ZL, YH:YL
	lpm	temp2, Z
	adiw	YH:YL, 1
	mov	ZL, temp2
	clr	ZH
	lsl	ZL
	rol	ZH
	lsl	ZL
	rol	ZH
	lsl	ZL
	rol	ZH
	add	ZL, temp2
	ldi	temp2, 0
	adc	ZH, temp2
	add	ZL, scanline
	adc	ZH, temp2
	subi	ZL, LOW(-charset*2)
	sbci	ZH, HIGH(-charset*2)
	lpm	temp2, Z
	lsl	temp2
	brcc	PC+2
	ori	temp, 0x02
	lsl	temp2
	brcc	PC+2
	ori	temp, 0x01
	out	PORTB, temp
	sbi	PORTD, LCD_CP ;CP high
	cbi	PORTD, LCD_CP ;CP low
	;char 3
	movw	ZH:ZL, YH:YL
	lpm	temp, Z
	adiw	YH:YL, 1
	mov	ZL, temp
	clr	ZH
	lsl	ZL
	rol	ZH
	lsl	ZL
	rol	ZH
	lsl	ZL
	rol	ZH
	add	ZL, temp
	ldi	temp, 0
	adc	ZH, temp
	add	ZL, scanline
	adc	ZH, temp
	subi	ZL, LOW(-charset*2)
	sbci	ZH, HIGH(-charset*2)
	lpm	temp, Z
	swap	temp
	mov	ZL, temp
	andi	ZL, 0x0F
	andi	temp, 0xC0
	or	temp2, ZL
	out	PORTB, temp2
	sbi	PORTD, LCD_CP ;CP high
	cbi	PORTD, LCD_CP ;CP low
	;char 4
	movw	ZH:ZL, YH:YL
	lpm	temp2, Z
	adiw	YH:YL, 1
	mov	ZL, temp2
	clr	ZH
	lsl	ZL
	rol	ZH
	lsl	ZL
	rol	ZH
	lsl	ZL
	rol	ZH
	add	ZL, temp2
	ldi	temp2, 0
	adc	ZH, temp2
	add	ZL, scanline
	adc	ZH, temp2
	subi	ZL, LOW(-charset*2)
	sbci	ZH, HIGH(-charset*2)
	lpm	temp2, Z
	lsr	temp2
	lsr	temp2
	or	temp2, temp
	out	PORTB, temp2
	sbi	PORTD, LCD_CP ;CP high
	cbi	PORTD, LCD_CP ;CP low
	;end of loop
	dec	count
	breq	PC+2
	rjmp	line_interrupt_text
	
	;shift the 21st char: pixels 121~128 (127 and 128 unused)
	movw	ZH:ZL, YH:YL
	lpm	temp, Z
	adiw	YH:YL, 1
	mov	ZL, temp
	clr	ZH
	lsl	ZL
	rol	ZH
	lsl	ZL
	rol	ZH
	lsl	ZL
	rol	ZH
	add	ZL, temp
	ldi	temp, 0
	adc	ZH, temp
	add	ZL, scanline
	adc	ZH, temp
	subi	ZL, LOW(-charset*2)
	sbci	ZH, HIGH(-charset*2)
	lpm	temp, Z
	out	PORTB, temp
	sbi	PORTD, LCD_CP ;CP high
	cbi	PORTD, LCD_CP ;CP low
	
	in	temp2, PORTD
	
.endif
	
	;latch data and switch to next line
	ori	temp2, 1<<LCD_LOAD ;LOAD/YSCLK high
	ldi	temp, 1<<LCD_LOAD
	cpi	line, 72
	brne	PC+3
	ori	temp, 1<<LCD_FR
	ori	temp2, 1<<LCD_DIO2 ;first line: DIO2 high
	out	PORTD, temp2
	eor	temp2, temp ;LOAD/YSCLK low, first line: toggle FR/M
	out	PORTD, temp2
	cbi	PORTD, LCD_DIO2 ;DIO2 low
	
.if (MODE == 0) ;image
	;line counter
	dec	line
	brne	PC+4
	ldi	line, 72
	ldi	ZL, LOW(image*2)
	ldi	ZH, HIGH(image*2)
.elseif (MODE == 1) ;text
	;scanline counter
	inc	scanline
	cpi	scanline, 9
	brne	PC+2
	ldi	scanline, 0
	breq	PC+3
	subi	YL, LOW(21)
	sbci	YH, HIGH(21)
	;line counter
	dec	line
	brne	PC+5
	ldi	line, 72
	ldi	YL, LOW(text*2)
	ldi	YH, HIGH(text*2)
	ldi	scanline, 0
.endif
	
	;return
	pop	count
	pop	temp2
	pop	temp
	out	SREG, sreg_backup
	reti
	
;===============================================================================

reset:
	;stackpointer
	ldi	temp, RAMEND
	out	SPL, temp
	
	;ports
	ldi	temp, 0xFF
	out	DDRB, temp
	ldi	temp, 0x00
	out	PORTB, temp
	ldi	temp, 0x7F
	out	DDRD, temp
	ldi	temp, 1<<LCD_ON
	out	PORTD, temp
	
	;init timer for LCD boost converter: 8 MHz / 256 = 31.25 kHz PWM
	ldi	temp, BOOST_PWM ;PWM value
	out	OCR0B, temp
	ldi	temp, 1<<COM0B1 | 1<<WGM01 | 1<<WGM00 ;fast PWM, non-inverted
	out	TCCR0A, temp
	ldi	temp, 0x01 ;Clk/1
	out	TCCR0B, temp
	
	;init timer for LCD line interrupt: 8 MHz / 8 / 139 / 72 = ~99.9 Hz
	ldi	temp2, HIGH(139-1) ;TOP value
	ldi	temp, LOW(139-1)
	out	OCR1AH, temp2
	out	OCR1AL, temp
	ldi	temp, 1<<WGM12 | 0x02 ;CTC mode, Clk/8
	out	TCCR1B, temp
	ldi	temp, 1<<OCIE1A
	out	TIMSK, temp
	
	;init registers
	ldi	line, 72
.if (MODE == 0) ;image
	ldi	ZL, LOW(image*2)
	ldi	ZH, HIGH(image*2)
.elseif (MODE == 1) ;text
	;text
	ldi	scanline, 0
	ldi	YL, LOW(text*2)
	ldi	YH, HIGH(text*2)
.endif
	
	;enable interrupts
	sei
	
;-------------------------------------------------------------------------------
	
loop:
	rjmp	loop
	
;===============================================================================

